// @flow
import _ from 'lodash';

export const actionTypes = {
	optimisticFailed : 'FetchAPI@OPTIMISTIC_FAILED',
	optimisticSuccess: 'FetchAPI@OPTIMISTIC_SUCCESS',
	optimisticUpdate : 'FetchAPI@OPTIMISTIC_UPDATE',
};

const
	createAction = ({dispatch, action}) => async ({
		body, callback, params, update, updateKey, validate = response => !_.isEmpty(response),
	}: {body: Object, callback: Function, params: Object, update: Function, updateKey: string, validate: Function}) => {
		let
			response = null;

		dispatch({type: actionTypes.optimisticUpdate, update, updateKey});

		try {
			response = await new Promise((resolve, reject) => dispatch(action(
				params,
				{body: JSON.stringify(body)},
				(err, res) => err ? reject(err) : resolve(res),
			)));
		} catch(err) {
			dispatch({type: actionTypes.optimisticFailed, updateKey});
			return callback(err, response);
		}

		// Custom validation method for use case like status === 200, but API RESULT is {result: 'Failed'}
		if(!validate(response)) {
			dispatch({response, type: actionTypes.optimisticFailed, updateKey});
			return callback(new Error('validation failed!!!'), response);
		}

		dispatch({response, type: actionTypes.optimisticSuccess, updateKey});
		callback(null, response);
		return response;
	};

export default (entityId: string, entity: Object) => _.extend(entity, {
	helpers: {
		optimisticDelete(props) {
			return createAction({
				action  : _.get(this.actions, `${entityId}.delete`),
				dispatch: this.dispatch,
				getState: this.getState,
			})(props);
		},

		optimisticPost(props) {
			return createAction({
				action  : _.get(this.actions, `${entityId}.post`),
				dispatch: this.dispatch,
				getState: this.getState,
			})(props);
		},

		optimisticPut(props) {
			return createAction({
				action  : _.get(this.actions, `${entityId}.put`),
				dispatch: this.dispatch,
				getState: this.getState,
			})(props);
		},
	},
});
